@@ -1,29 +1,41 @@
 use std::collections::HashSet;
 use std::path::{Path, PathBuf};
 use std::process::Stdio;
+use std::str::Utf8Error;
 use std::sync::LazyLock;
 
 use anyhow::Result;
 use itertools::Itertools;
+use rustc_hash::FxHashSet;
 use tokio::io::AsyncWriteExt;
-use tracing::warn;
+use tracing::{debug, warn};
 
+use crate::process;
 use crate::process::{Cmd, StatusError};
-use crate::{git, process};
 
 #[derive(Debug, thiserror::Error)]
 pub(crate) enum Error {
     #[error(transparent)]
     Command(#[from] process::Error),
+
     #[error("Failed to find git: {0}")]
     GitNotFound(#[from] which::Error),
+
     #[error(transparent)]
     Io(#[from] std::io::Error),
+
+    #[error(transparent)]
+    UTF8(#[from] Utf8Error),
 }
 
 pub(crate) static GIT: LazyLock<Result<PathBuf, which::Error>> =
     LazyLock::new(|| which::which("git"));
-pub(crate) static GIT_ROOT: LazyLock<Result<PathBuf, Error>> = LazyLock::new(get_root);
+
+pub(crate) static GIT_ROOT: LazyLock<Result<PathBuf, Error>> = LazyLock::new(|| {
+    get_root().inspect(|root| {
+        debug!("Git root: {}", root.display());
+    })
+});
 
 // Remove some `GIT_` environment variables exposed by `git`.
 
@@ -61,19 +73,14 @@ pub(crate) fn git_cmd(summary: &str) -> Result<Cmd, Error> {
     Ok(cmd)
 }
 
-fn zsplit(s: &[u8]) -> Vec<String> {
+fn zsplit(s: &[u8]) -> Result<Vec<PathBuf>, Utf8Error> {
     s.split(|&b| b == b'\0')
-        .filter_map(|slice| {
-            if slice.is_empty() {
-                None
-            } else {
-                Some(String::from_utf8_lossy(slice).to_string())
-            }
-        })
+        .filter(|slice| !slice.is_empty())
+        .map(|slice| str::from_utf8(slice).map(PathBuf::from))
         .collect()
 }
 
-pub(crate) async fn intent_to_add_files() -> Result<Vec<String>, Error> {
+pub(crate) async fn intent_to_add_files() -> Result<Vec<PathBuf>, Error> {
     let output = git_cmd("get intent to add files")?
         .arg("diff")
         .arg("--no-ext-diff")
@@ -84,10 +91,10 @@ pub(crate) async fn intent_to_add_files() -> Result<Vec<String>, Error> {
         .check(true)
         .output()
         .await?;
-    Ok(zsplit(&output.stdout))
+    Ok(zsplit(&output.stdout)?)
 }
 
-pub(crate) async fn get_changed_files(old: &str, new: &str) -> Result<Vec<String>, Error> {
+pub(crate) async fn get_changed_files(old: &str, new: &str) -> Result<Vec<PathBuf>, Error> {
     let output = git_cmd("get changed files")?
         .arg("diff")
         .arg("--name-only")
@@ -98,18 +105,18 @@ pub(crate) async fn get_changed_files(old: &str, new: &str) -> Result<Vec<String
         .check(true)
         .output()
         .await?;
-    Ok(zsplit(&output.stdout))
+    Ok(zsplit(&output.stdout)?)
 }
 
-pub(crate) async fn git_ls_files(path: Option<&Path>) -> Result<Vec<String>, Error> {
+pub(crate) async fn ls_files(cwd: &Path, path: Option<&Path>) -> Result<Vec<PathBuf>, Error> {
     let mut cmd = git_cmd("get git all files")?;
-    cmd.arg("ls-files").arg("-z").check(true);
+    cmd.current_dir(cwd).arg("ls-files").arg("-z").check(true);
 
     if let Some(p) = path {
         cmd.arg("--").arg(p);
     }
     let output = cmd.output().await?;
-    Ok(zsplit(&output.stdout))
+    Ok(zsplit(&output.stdout)?)
 }
 
 pub(crate) async fn get_git_dir() -> Result<PathBuf, Error> {
@@ -140,7 +147,7 @@ pub(crate) async fn get_git_common_dir() -> Result<PathBuf, Error> {
     }
 }
 
-pub(crate) async fn get_staged_files() -> Result<Vec<String>, Error> {
+pub(crate) async fn get_staged_files() -> Result<Vec<PathBuf>, Error> {
     let output = git_cmd("get staged files")?
         .arg("diff")
         .arg("--staged")
@@ -151,22 +158,26 @@ pub(crate) async fn get_staged_files() -> Result<Vec<String>, Error> {
         .check(true)
         .output()
         .await?;
-    Ok(zsplit(&output.stdout))
+    Ok(zsplit(&output.stdout)?)
 }
 
-pub(crate) async fn file_not_staged(file: &Path) -> Result<bool> {
-    let status = git::git_cmd("git diff")?
+pub(crate) async fn files_not_staged(files: &[&Path]) -> Result<Vec<PathBuf>> {
+    let output = git_cmd("git diff")?
         .arg("diff")
-        .arg("--quiet") // Implies --exit-code
+        .arg("--exit-code")
+        .arg("--name-only")
         .arg("--no-ext-diff")
-        .arg(file)
-        .stdout(Stdio::null())
-        .stderr(Stdio::null())
+        .arg("-z") // Use NUL as line terminator
+        .args(files)
         .check(false)
-        .status()
+        .output()
         .await?;
 
-    Ok(status.code().is_some_and(|code| code == 1))
+    if output.status.code().is_some_and(|code| code == 1) {
+        return Ok(zsplit(&output.stdout)?);
+    }
+
+    Ok(vec![])
 }
 
 pub(crate) async fn has_unmerged_paths() -> Result<bool, Error> {
@@ -184,7 +195,7 @@ pub(crate) async fn is_in_merge_conflict() -> Result<bool, Error> {
     Ok(git_dir.join("MERGE_HEAD").try_exists()? && git_dir.join("MERGE_MSG").try_exists()?)
 }
 
-pub(crate) async fn get_conflicted_files() -> Result<Vec<String>, Error> {
+pub(crate) async fn get_conflicted_files() -> Result<Vec<PathBuf>, Error> {
     let tree = git_cmd("git write-tree")?
         .arg("write-tree")
         .check(true)
@@ -196,7 +207,7 @@ pub(crate) async fn get_conflicted_files() -> Result<Vec<String>, Error> {
         .arg("--name-only")
         .arg("--no-ext-diff") // Disable external diff drivers
         .arg("-z") // Use NUL as line terminator
-        .arg("-m")
+        .arg("-m") // Show diffs for merge commits in the default format.
         .arg(String::from_utf8_lossy(&tree.stdout).trim())
         .arg("HEAD")
         .arg("MERGE_HEAD")
@@ -204,33 +215,37 @@ pub(crate) async fn get_conflicted_files() -> Result<Vec<String>, Error> {
         .output()
         .await?;
 
-    Ok(zsplit(&output.stdout)
+    Ok(zsplit(&output.stdout)?
         .into_iter()
         .chain(parse_merge_msg_for_conflicts().await?)
-        .collect::<HashSet<String>>()
+        .collect::<HashSet<PathBuf>>()
         .into_iter()
         .collect())
 }
 
-async fn parse_merge_msg_for_conflicts() -> Result<Vec<String>, Error> {
+async fn parse_merge_msg_for_conflicts() -> Result<Vec<PathBuf>, Error> {
     let git_dir = get_git_dir().await?;
     let merge_msg = git_dir.join("MERGE_MSG");
-    let content = fs_err::read_to_string(&merge_msg)?;
+    let content = fs_err::tokio::read_to_string(&merge_msg).await?;
     let conflicts = content
         .lines()
         // Conflicted files start with tabs
         .filter(|line| line.starts_with('\t') || line.starts_with("#\t"))
         .map(|line| line.trim_start_matches('#').trim().to_string())
+        .map(PathBuf::from)
         .collect();
+
     Ok(conflicts)
 }
 
-pub(crate) async fn get_diff() -> Result<Vec<u8>, Error> {
+pub(crate) async fn get_diff(path: &Path) -> Result<Vec<u8>, Error> {
     let output = git_cmd("git diff")?
         .arg("diff")
         .arg("--no-ext-diff") // Disable external diff drivers
         .arg("--no-textconv")
         .arg("--ignore-submodules")
+        .arg("--")
+        .arg(path)
         .check(true)
         .output()
         .await?;
@@ -252,7 +267,8 @@ pub(crate) async fn write_tree() -> Result<String, Error> {
 
 /// Get the path of the top-level directory of the working tree.
 pub(crate) fn get_root() -> Result<PathBuf, Error> {
-    let output = std::process::Command::new(GIT.clone()?)
+    let git = GIT.as_ref().map_err(|&e| Error::GitNotFound(e))?;
+    let output = std::process::Command::new(git)
         .arg("rev-parse")
         .arg("--show-toplevel")
         .output()?;
@@ -261,7 +277,7 @@ pub(crate) fn get_root() -> Result<PathBuf, Error> {
             summary: "get git root".to_string(),
             error: StatusError {
                 status: output.status,
-                output: None,
+                output: Some(output),
             },
         }));
     }
@@ -271,25 +287,6 @@ pub(crate) fn get_root() -> Result<PathBuf, Error> {
     ))
 }
 
-pub(crate) async fn is_dirty(path: &Path) -> Result<bool, Error> {
-    let mut cmd = git_cmd("check git is dirty")?;
-    let output = cmd
-        .arg("diff")
-        .arg("--quiet") // Implies `--exit-code`
-        .arg("--no-ext-diff") // Disable external diff drivers
-        .arg(path)
-        .check(false)
-        .output()
-        .await?;
-    if output.status.success() {
-        Ok(false)
-    } else if output.status.code() == Some(1) {
-        Ok(true)
-    } else {
-        Err(cmd.check_status(output.status).unwrap_err().into())
-    }
-}
-
 pub(crate) async fn init_repo(url: &str, path: &Path) -> Result<(), Error> {
     git_cmd("init git repo")?
         .arg("init")
@@ -414,7 +411,7 @@ pub(crate) async fn has_hooks_path_set() -> Result<bool> {
     }
 }
 
-pub(crate) async fn lfs_files<T: FromIterator<String>>(paths: &[&String]) -> Result<T, Error> {
+pub(crate) async fn get_lfs_files(paths: &[&Path]) -> Result<Vec<String>, Error> {
     let mut job = git_cmd("git check-attr")?
         .arg("check-attr")
         .arg("filter")
@@ -423,12 +420,19 @@ pub(crate) async fn lfs_files<T: FromIterator<String>>(paths: &[&String]) -> Res
         .stdin(Stdio::piped())
         .stdout(Stdio::piped())
         .check(true)
-        // .output()
         .spawn()?;
 
     {
         let mut stdin = job.stdin.take().expect("Failed to open stdin");
-        stdin.write_all(paths.iter().join("\0").as_ref()).await?;
+        stdin
+            .write_all(
+                paths
+                    .iter()
+                    .map(|f| f.to_string_lossy())
+                    .join("\0")
+                    .as_ref(),
+            )
+            .await?;
     }
 
     Ok(
@@ -446,3 +450,70 @@ pub(crate) async fn lfs_files<T: FromIterator<String>>(paths: &[&String]) -> Res
             .collect(),
     )
 }
+
+/// Check if a git revision exists
+pub(crate) async fn rev_exists(rev: &str) -> Result<bool, Error> {
+    let output = git_cmd("check if revision exists")?
+        .arg("rev-list")
+        .arg("--quiet")
+        .arg(rev)
+        .check(false)
+        .output()
+        .await?;
+    Ok(output.status.success())
+}
+
+/// Get commits that are ancestors of the given commit but not in the specified remote
+pub(crate) async fn get_ancestors_not_in_remote(
+    local_sha: &str,
+    remote_name: &str,
+) -> Result<Vec<String>, Error> {
+    let output = git_cmd("get ancestors not in remote")?
+        .arg("rev-list")
+        .arg(local_sha)
+        .arg("--topo-order")
+        .arg("--reverse")
+        .arg("--not")
+        .arg(format!("--remotes={remote_name}"))
+        .check(true)
+        .output()
+        .await?;
+    Ok(String::from_utf8_lossy(&output.stdout)
+        .trim()
+        .lines()
+        .map(ToString::to_string)
+        .collect())
+}
+
+/// Get root commits (commits with no parents) for the given commit
+pub(crate) async fn get_root_commits(local_sha: &str) -> Result<FxHashSet<String>, Error> {
+    let output = git_cmd("get root commits")?
+        .arg("rev-list")
+        .arg("--max-parents=0")
+        .arg(local_sha)
+        .check(true)
+        .output()
+        .await?;
+    Ok(String::from_utf8_lossy(&output.stdout)
+        .trim()
+        .lines()
+        .map(ToString::to_string)
+        .collect())
+}
+
+/// Get the parent commit of the given commit
+pub(crate) async fn get_parent_commit(commit: &str) -> Result<Option<String>, Error> {
+    let output = git_cmd("get parent commit")?
+        .arg("rev-parse")
+        .arg(format!("{commit}^"))
+        .check(false)
+        .output()
+        .await?;
+    if output.status.success() {
+        Ok(Some(
+            String::from_utf8_lossy(&output.stdout).trim().to_string(),
+        ))
+    } else {
+        Ok(None)
+    }
+}
