@@ -7,19 +7,21 @@ use std::sync::{Arc, OnceLock};
 
 use anyhow::{Context, Result};
 use clap::ValueEnum;
+use constants::MANIFEST_FILE;
 use rand::Rng;
 use rustc_hash::{FxBuildHasher, FxHashMap, FxHashSet};
 use serde::{Deserialize, Serialize};
 use thiserror::Error;
-use tracing::{error, warn};
+use tracing::{error, trace};
 
 use crate::config::{
-    self, Config, HookOptions, Language, LocalHook, MANIFEST_FILE, ManifestHook, MetaHook,
-    RemoteHook, SerdeRegex, Stage, read_manifest,
+    self, Config, HookOptions, Language, LocalHook, ManifestHook, MetaHook, RemoteHook, SerdeRegex,
+    Stage, read_manifest,
 };
 use crate::languages::version::LanguageRequest;
 use crate::languages::{extract_metadata_from_entry, resolve_command};
 use crate::store::Store;
+use crate::workspace::Project;
 
 #[derive(Error, Debug)]
 pub(crate) enum Error {
@@ -117,6 +119,7 @@ impl Display for Repo {
 }
 
 pub(crate) struct HookBuilder {
+    project: Arc<Project>,
     repo: Arc<Repo>,
     config: ManifestHook,
     // The index of the hook in the project configuration.
@@ -124,8 +127,18 @@ pub(crate) struct HookBuilder {
 }
 
 impl HookBuilder {
-    pub(crate) fn new(repo: Arc<Repo>, config: ManifestHook, idx: usize) -> Self {
-        Self { repo, config, idx }
+    pub(crate) fn new(
+        project: Arc<Project>,
+        repo: Arc<Repo>,
+        config: ManifestHook,
+        idx: usize,
+    ) -> Self {
+        Self {
+            project,
+            repo,
+            config,
+            idx,
+        }
     }
 
     /// Update the hook from the project level hook configuration.
@@ -270,7 +283,7 @@ impl HookBuilder {
             language_request,
             additional_dependencies,
             dependencies: OnceLock::new(),
-
+            project: self.project,
             repo: self.repo,
             idx: self.idx,
             id: self.config.id,
@@ -294,7 +307,12 @@ impl HookBuilder {
         };
 
         if let Err(err) = extract_metadata_from_entry(&mut hook).await {
-            warn!("Failed to extract metadata from entry for hook `{hook}`: {err}");
+            if err
+                .downcast_ref::<std::io::Error>()
+                .is_some_and(|e| e.kind() != std::io::ErrorKind::NotFound)
+            {
+                trace!("Failed to extract metadata from entry for hook `{hook}`: {err}");
+            }
         }
 
         Ok(hook)
@@ -367,6 +385,7 @@ impl Entry {
 #[allow(clippy::struct_excessive_bools)]
 #[derive(Debug, Clone)]
 pub(crate) struct Hook {
+    project: Arc<Project>,
     repo: Arc<Repo>,
     // Cached computed dependencies.
     dependencies: OnceLock<FxHashSet<String>>,
@@ -408,6 +427,10 @@ impl Display for Hook {
 }
 
 impl Hook {
+    pub(crate) fn project(&self) -> &Project {
+        &self.project
+    }
+
     pub(crate) fn repo(&self) -> &Repo {
         &self.repo
     }
@@ -417,6 +440,20 @@ impl Hook {
         self.repo.path()
     }
 
+    pub(crate) fn full_id(&self) -> String {
+        let path = self.project.relative_path();
+        if path.as_os_str().is_empty() {
+            format!(".:{}", self.id)
+        } else {
+            format!("{}:{}", path.display(), self.id)
+        }
+    }
+
+    /// Get the path where the hook should be executed.
+    pub(crate) fn work_dir(&self) -> &Path {
+        self.project.path()
+    }
+
     pub(crate) fn is_local(&self) -> bool {
         matches!(&*self.repo, Repo::Local { .. })
     }
@@ -474,6 +511,7 @@ impl Display for InstalledHook {
 }
 
 impl InstalledHook {
+    /// Get the path to the environment where the hook is installed.
     pub(crate) fn env_path(&self) -> Option<&Path> {
         match self {
             InstalledHook::Installed { info, .. } => Some(&info.env_path),
@@ -571,7 +609,7 @@ impl InstallInfo {
 
     pub fn matches(&self, hook: &Hook) -> bool {
         self.language == hook.language
-            && self.dependencies.is_superset(hook.dependencies())
+            && &self.dependencies == hook.dependencies()
             && hook.language_request.satisfied_by(self)
     }
 }
