@@ -1,3 +1,4 @@
+use std::path::Path;
 use std::sync::Arc;
 
 use anyhow::Result;
@@ -29,13 +30,14 @@ impl LanguageImpl for DockerImage {
     async fn run(
         &self,
         hook: &InstalledHook,
-        filenames: &[&String],
+        filenames: &[&Path],
         _store: &Store,
     ) -> Result<(i32, Vec<u8>)> {
         let entry = hook.entry.resolve(None)?;
-        let run = async move |batch: &[&String]| {
-            let mut cmd = Docker::docker_run_cmd()?;
+        let run = async move |batch: &[&Path]| {
+            let mut cmd = Docker::docker_run_cmd(hook.work_dir())?;
             let mut output = cmd
+                .current_dir(hook.work_dir())
                 .args(&entry[..])
                 .args(&hook.args)
                 .args(batch)
