@@ -2,8 +2,6 @@ use assert_fs::assert::PathAssert;
 use assert_fs::fixture::PathChild;
 
 use crate::common::{TestContext, cmd_snapshot};
-
-// We use `setup-go` action to install go1.24.5 in CI, so 1.23.11 should be downloaded by prek.
 #[test]
 fn language_version() -> anyhow::Result<()> {
     let context = TestContext::new();
@@ -102,17 +100,19 @@ fn language_version() -> anyhow::Result<()> {
     ----- stderr -----
     "#);
 
-    assert_eq!(
-        context
-            .home_dir()
-            .join("tools")
-            .join("go")
-            .read_dir()?
-            .flatten()
-            .filter(|d| !d.file_name().to_string_lossy().starts_with('.'))
-            .map(|d| d.file_name().to_string_lossy().to_string())
-            .collect::<Vec<_>>(),
-        vec!["1.23.11"],
+    let installed_versions = context
+        .home_dir()
+        .join("tools")
+        .join("go")
+        .read_dir()?
+        .flatten()
+        .filter(|d| !d.file_name().to_string_lossy().starts_with('.'))
+        .map(|d| d.file_name().to_string_lossy().to_string())
+        .collect::<Vec<_>>();
+
+    assert!(
+        installed_versions.contains(&"1.23.11".to_string()),
+        "Go version 1.23.11 not found in installed versions: {installed_versions:?}",
     );
 
     Ok(())
@@ -160,7 +160,7 @@ fn remote_hook() {
       	-lang       str    target Go version in the form "go1.X" (default from go.mod)
       	-modpath    str    Go module path containing the source file (default from go.mod)
 
-    ----- stderr -----
+    ----- stderr ----- 
     "#);
 
     // Run hooks with newly downloaded go.
@@ -184,7 +184,7 @@ fn remote_hook() {
     - duration: [TIME]
       .pre-commit-config.yaml
 
-    ----- stderr -----
+    ----- stderr ----- 
     ");
 
     // Run hooks with system found go.
@@ -208,6 +208,6 @@ fn remote_hook() {
     - duration: [TIME]
       .pre-commit-config.yaml
 
-    ----- stderr -----
+    ----- stderr ----- 
     ");
 }
